import gradio_client as gc
import functools
from typing import List
from .prompt_templates import ChatAgent, RAGAgent
from agentx.engine_websocket import generate

END_OF_MESSAGE_TURN_HUMAN = "<|END_OF_MESSAGE_TURN_HUMAN|>"
END_OF_MESSAGE = "<|END_OF_MESSAGE|>"


def format_chat_for_ai_client(user: List[str], assistance: List[str]):
    history = ""
    for c1, c2 in zip(user, assistance):
        history += f"{c1}{END_OF_MESSAGE_TURN_HUMAN}{c2}{END_OF_MESSAGE}"
    return history


class PixelClient:
    def __init__(
            self,
            hostname: str,
    ):
        assert not hostname.startswith("http://") and not hostname.startswith("https://"), "Invalid hostname format"
        self.hostname = hostname

    def __call__(
            self,
            prompt: str,
            conversation_history: List[dict] = None,
            contexts: List[str] = None,
            debug: bool = False,
            max_new_tokens: int = 2048,
            stream: bool = False
    ) -> str:
        """

        :param prompt:string
        :param conversation_history:[{"user": ..., "assistance": ..., ...}]
        :param contexts: List of Contexts [str]
        :param debug: Bool
        :param max_new_tokens: int
        :return: response
        """

        conversation = None

        if contexts is not None:
            if conversation_history is not None:
                conversation = []
                for user, assistant in zip(
                        [f["user"] for f in conversation_history],
                        [f["assistance"] for f in conversation_history]
                ):
                    conversation.append(
                        [user, assistant]
                    )
            prompt = RAGAgent.render(
                prompt=prompt,
                conversation=conversation,
                contexts=contexts
            )
        else:
            if conversation_history is not None:
                conversation = []
                for user, assistant in zip(
                        [f["user"] for f in conversation_history],
                        [f["assistance"] for f in conversation_history]
                ):
                    conversation.append(
                        [user, assistant]
                    )

            prompt = ChatAgent.render(
                prompt=prompt,
                conversation=conversation
            )
        assert isinstance(stream, bool), "stream must be bool"
        total_response = ""
        for response in generate(
                hostname=self.hostname,
                prompt=prompt,
                stream=True
        ):
            if stream:
                yield response.response
            total_response += response.response
        if not stream:
            return total_response



from typing import Optional
from jinja2 import Environment, BaseLoader
import os

PROMPT = open(f"{os.path.dirname(__file__)}/prompt.jinja2", "r").read().strip()


class RAGAgent:
    def __init__(
            self,
            **kwargs
    ):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @staticmethod
    def render(
            prompt: str,
            contexts: list[str],
            conversation: Optional[list[list[str]]] = None,
    ) -> str:
        template = Environment(loader=BaseLoader()).from_string(PROMPT)
        return template.render(
            conversation=conversation,
            prompt=prompt,
            contexts=contexts
        )



import gradio as gr

from agentx import ServeEngine
import logging
from typing import List, Optional
import transformers

logging.basicConfig(
    level=logging.INFO
)


class AgentXServer(ServeEngine):
    tokenizer: transformers.PreTrainedTokenizerBase | None = None

    def create_gradio_pixely_ai(self):
        with gr.Blocks(theme=gr.themes.Soft()) as block:
            gr.Markdown("# <h3> <center>Powered by [AgentX](https://github.com/erfanzar/AgentX) </center> </h3>")

            with gr.Row():
                with gr.Column():
                    prompt = gr.Textbox(
                        show_label=True,
                        placeholder="Message Box",
                        container=True,
                        label="Message Box"
                    )
                    system = gr.Textbox(
                        show_label=True,
                        placeholder="System",
                        container=True,
                        value="",
                        label="System"
                    )
                    response = gr.Markdown(
                        show_label=True,
                        label="Response"
                    )
                    submit = gr.Button(variant="primary")
            with gr.Row():
                with gr.Accordion("Advanced Options", open=False):
                    max_new_tokens = gr.Slider(
                        value=self.sample_config.max_new_tokens,
                        maximum=10000,
                        minimum=1,
                        label="Max New Tokens",
                        step=1
                    )

            inputs = [
                prompt,
                system,
                max_new_tokens,
            ]
            _ = prompt.submit(
                fn=self.process_pixely_request,
                inputs=inputs,
                outputs=[prompt, response]
            )
            _ = submit.click(
                fn=self.process_pixely_request,
                inputs=inputs,
                outputs=[prompt, response]
            )

            block.queue()
        return block

    def process_pixely_request(
            self,
            prompt: str,
            system: str,
            max_new_tokens: int,
    ):
        conversation = []

        if system is not None and system != "":
            conversation.append({"role": "system", "content": system})
        conversation.append({"role": "user", "content": prompt})
        if self.prompt_template is None:
            prompt = self.tokenizer.apply_chat_template(
                conversation,
                tokenize=False,
                add_generation_prompt=True
            )
        else:
            prompt = self.prompt_template.render(conversation)
        response = ""
        for char in self.process(
                prompt=prompt,
                max_new_tokens=max_new_tokens,
                max_sequence_length=max_new_tokens
        ):
            response += char
            yield "", response

    def build(
            self,
    ):
        with gr.Blocks(
                theme=gr.themes.Soft(
                    primary_hue=gr.themes.colors.orange,
                    secondary_hue=gr.themes.colors.orange,
                ),
                title="AgentX inference"
        ) as block:
            with gr.Tab("Pixely Serve"):
                self.create_gradio_pixely_ai()
            with gr.Tab("Pixely Chat"):
                self.chat_interface_components()
        return block



from .serving import AgentXServer as AgentXServer
from .client import PixelClient as PixelClient
from .prompt_templates import (
    ChatAgent as ChatAgent,
    RAGAgent as RAGAgent
)

__all__ = "PixelClient",
__version__ = "0.0.41"



import asyncio
import websockets

import json


async def handle_client(websocket, path):
    try:
        # Receive message from client
        message = await websocket.recv()
        data = json.loads(message)  # Parse JSON data

        # Process the dict with 'prompt' and 'conversation'
        prompt = data.get('prompt')
        conversation = data.get('conversation')  # This should be a list of dicts

        # Here you would process the input and generate tokens (mocked as a response)
        response_data = {
            'tokens': ['token1', 'token2', 'token3']  # Replace with actual token generation logic
        }

        # Send the response back to the client
        response = json.dumps(response_data)
        await websocket.send(response)
    except websockets.ConnectionClosed:
        print("Connection closed")
    except Exception as e:
        print(f"Error: {e}")


if __name__ == '__main__':
    # Start the server
    start_server = websockets.serve(handle_client, "localhost", 6789)

    asyncio.get_event_loop().run_until_complete(start_server)
    asyncio.get_event_loop().run_forever()
